/******************************************************************************
 * shadowmap.cpp 
 * implements the construction and drawing of shadow maps
 * Copyright (C) 2010 Marc Gissler <gisslerm@informatik.uni-freiburg.de>
 ******************************************************************************/

#include "shadowmap.h"

#include "shaderProgramDepth.h"
#include "shaderProgramShadowMap.h"

#include "lightmanager.h"
#include "scene.h"

CShadowMap::CShadowMap() : m_shaderShadowMap(NULL), m_shaderDepth(NULL)
{
	m_shadowMapSize[0] = 1024;
	m_shadowMapSize[1] = 1024;

	// build the necessary shaders
	m_shaderDepth = new CShaderProgramDepth();
	if(!m_shaderDepth->isValid())
	{
		delete m_shaderDepth;
		printf("depth shader is not valid!");
		exit(-1);
	}

	m_shaderShadowMap = new CShaderProgramShadowMap();
	if(!m_shaderShadowMap->isValid())
	{
		delete m_shaderShadowMap;
		printf("shadow map shader is not valid!");
		exit(-1);
	}

	// prepare offscreen rendering

	// prepare shadow map textures
	glGenTextures(1, &m_shadowMap);
	glBindTexture(GL_TEXTURE_2D, m_shadowMap);
	glTexParameteri(GL_TEXTURE_2D, GL_DEPTH_TEXTURE_MODE, GL_LUMINANCE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_MODE, GL_COMPARE_R_TO_TEXTURE);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_FUNC, GL_LEQUAL);
	glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP/*GL_CLAMP_TO_EDGE*/);
	glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP/*GL_CLAMP_TO_EDGE*/);
	//glTexImage2D(GL_TEXTURE_2D, 0, GL_DEPTH24_STENCIL8, m_shadowMapSize, m_shadowMapSize, 0, GL_DEPTH_STENCIL, GL_UNSIGNED_INT_24_8, NULL);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_DEPTH_COMPONENT, m_shadowMapSize[0], m_shadowMapSize[1], 0, GL_DEPTH_COMPONENT, GL_FLOAT, NULL);

	glGenTextures(1, &m_shadowMapColor);
	glBindTexture(GL_TEXTURE_2D, m_shadowMapColor);
	glTexImage2D(GL_TEXTURE_2D, 0, /*GL_RGBA*/GL_RGBA32F_ARB,  m_shadowMapSize[0], m_shadowMapSize[1], 0, GL_RGB, GL_FLOAT, NULL);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP/*GL_CLAMP_TO_EDGE*/);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP/*GL_CLAMP_TO_EDGE*/);

	//create frame buffer
	glGenFramebuffersEXT(1, &m_shadowFrameBufferHandle);
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, m_shadowFrameBufferHandle);
	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT, GL_TEXTURE_2D, m_shadowMap, 0);
	//glFramebufferTexture2DEXT(GL_FRAMEBUFFER, GL_STENCIL_ATTACHMENT, GL_TEXTURE_2D, m_shadowMap, 0);
	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, m_shadowMapColor, 0);

	// if the frame buffers could not be created, exit the program
	if(!m_checkValidity()) 
	{
		glDeleteTextures(1, &m_shadowMap);			
		glDeleteTextures(1, &m_shadowMapColor);			
		glDeleteFramebuffersEXT(1,  &m_shadowFrameBufferHandle);
		m_shadowMap = m_shadowMapColor = m_shadowFrameBufferHandle = 0;
		exit(-1);
	}
	
	// unbind, for now (to get back to the original frame buffer)
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
}

CShadowMap::~CShadowMap()
{
	glDeleteTextures(1, &m_shadowMap);
	glDeleteTextures(1, &m_shadowMapColor);

	glDeleteFramebuffersEXT(1,  &m_shadowFrameBufferHandle);
}

void CShadowMap::build(int lightId, gmVector3& lightPos)
{
	std::map<int, CShadowData*>::iterator mapIter;
	mapIter = m_mapLightAndData.find(lightId);

	if(mapIter == m_mapLightAndData.end())
	{
		CShadowData* data = new CShadowData();
		mapIter = m_mapLightAndData.insert(std::pair<int, CShadowData*>(lightId, data)).first;
	}
	//CShadowData* data = /*static_cast<CShadowData*>(*/(*mapIter).second/*)*/;
}

void CShadowMap::clear()
{
	std::map<int, CShadowData*>::iterator mapIter;
	for(mapIter = m_mapLightAndData.begin(); mapIter != m_mapLightAndData.end(); ++mapIter)
	{
		CShadowData* data = /*static_cast<CShadowData*>(*/mapIter->second/*)*/;
		delete data;
	}
	m_mapLightAndData.clear();
}

bool CShadowMap::m_checkValidity()
{
	bool success = true;

	GLuint error = glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
	switch(error)
	{
	case GL_FRAMEBUFFER_COMPLETE_EXT:
		{
			printf("FrameBufferObjects initialized.");
			break;
		}
	case GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT_EXT:
		{
			success = false;
			printf("Incomplete attachment in FBO");
			break;
		}
	case GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT_EXT:
		{
			success = false;
			printf("Missing attachment in FBO");
			break;
		}
	case GL_FRAMEBUFFER_INCOMPLETE_DIMENSIONS_EXT:
		{
			success = false;
			printf("Incomplete dimensions");
			break;
		}
	case GL_FRAMEBUFFER_INCOMPLETE_FORMATS_EXT:
		{
			success = false;
			printf("Incomplete formats");
			break;
		}
	case GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER_EXT:
		{
			success = false;
			printf("Incomplete draw buffer");
			break;
		}
	case GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER_EXT:
		{
			success = false;
			printf("Incomplete read buffer");
			break;
		}
	case GL_FRAMEBUFFER_UNSUPPORTED_EXT:
		{
			success = false;
			printf("Framebufferobjects unsupported");
			break;
		}
	}

	return success;
}

void CShadowMap::drawSceneWithShadows()
{
	// 1. pass draw scene as usual

	CScene* pScene = CScene::getScene();
	std::map<int, CShadowData*>::iterator mapIter;

	/*for(mapIter = m_mapLightAndData.begin(); mapIter != m_mapLightAndData.end(); ++mapIter)
	{
		int lightId = mapIter->first;
		glEnable(GL_LIGHT0+lightId);
	}*/

	//TODO: render ambient intensities (for now, we render all intensities)
	// render ambient intensities
	//CLightManager::getLightManager()->setAmbientExclusively();
	
//	glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);

	glEnable(GL_CULL_FACE);
	glCullFace(GL_BACK);
	
	pScene->drawMeshes();
	/*CLightManager::getLightManager()->setAllIntensities();*/

//	glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);

	/*for(mapIter = m_mapLightAndData.begin(); mapIter != m_mapLightAndData.end(); ++mapIter)
	{
		int lightId = mapIter->first;
		glDisable(GL_LIGHT0+lightId);
	}*/

	//debug
	/*{
		return;
	}*/

	// render shadow maps and shadows for each of the light sources 
	for(mapIter = m_mapLightAndData.begin(); mapIter != m_mapLightAndData.end(); ++mapIter)
	{
		int lightId = mapIter->first;
		//glEnable(GL_LIGHT0+lightId); // lighting not needed
		CLight* pLight = CLightManager::getLightManager()->getLight(lightId);
		
		// compute the shadow map (using the depth shader)

		// ----------------------------------------------------------------------------------------------------
		// step 1: draw the shadow map
		//

		// bind the frame buffer (from here on, we render into an offscreen buffer, until we unbind the frame buffer again)
		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, m_shadowFrameBufferHandle);
		
		glClearColor(Globals::clearColor[0], Globals::clearColor[1], Globals::clearColor[2], 0.0f);
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		glEnable(GL_CULL_FACE);
		glCullFace(GL_FRONT);

		// save current viewport
		int viewport[4];
		glGetIntegerv(GL_VIEWPORT, (int*) viewport);

		// set viewport and render scene
		glViewport(0,0, m_shadowMapSize[0], m_shadowMapSize[1]);
		
		glMatrixMode(GL_PROJECTION);
		glPushMatrix();
		glLoadIdentity();
		
		// see also reshape() in viewer.cpp (should be consistent!)
		double znear = 1.0;													// position of near clipping plane
		double zfar = 999.0;													// position of far clipping plane
		double aspectRatio = m_shadowMapSize[0] / m_shadowMapSize[1];	// aspect ratio
		gluPerspective(65, aspectRatio, znear, zfar);

		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();
		glLoadIdentity();

		// get a light's position and look at the center of the scene (we assume this to be (0, 0, 0) !)
		gluLookAt(pLight->m_pos[0], pLight->m_pos[1], pLight->m_pos[2],
			0.0, 0.0, 0.0,
			0.0, 1.0, 0.0);

		// use the depth shader in this render pass (results is stored in m_shadowMapColor texture handle)
		m_shaderDepth->use();
		pScene->drawMeshes();

		//glDisable(GL_CULL_FACE);
		glCullFace(GL_BACK);

		//restore viewport
		glViewport(viewport[0],viewport[1],viewport[2],viewport[3]);

		glMatrixMode(GL_PROJECTION);
		glPopMatrix();

		glMatrixMode(GL_MODELVIEW);
		glPopMatrix();
		
		////debug
		//{
		//	int size = m_shadowMapSize * m_shadowMapSize * sizeof(GL_FLOAT);
		//	GLfloat *bytes = (GLfloat*) malloc( size);
		//	glBindTexture(GL_TEXTURE_2D, m_shadowMapColor);
		//	glGetTexImage(GL_TEXTURE_2D, 0, /*GL_RGBA*/ GL_LUMINANCE, GL_FLOAT, bytes);
		//	std::vector<float> vect;
		//	vect.reserve(size);
		//	vect.resize(size);
		//	for(size_t i = 0; i < size/ sizeof(GL_FLOAT); ++i)
		//		vect[i] = bytes[i];
		//	free(bytes);

		////	glDrawBuffer(GL_COLOR_ATTACHMENT0);
		////	glReadBuffer(GL_COLOR_ATTACHMENT0);
		//}

		// ----------------------------------------------------------------------------------------------------
		// step 2: set texture matrix 7
		//
		static double modelView[16];
		static double projection[16];

		// moving from unit cube [-1,1] to [0,1]  
		const GLdouble bias[16] = {	
			0.5, 0.0, 0.0, 0.0, 
			0.0, 0.5, 0.0, 0.0,
			0.0, 0.0, 0.5, 0.0,
			0.5, 0.5, 0.5, 1.0
		};

		glMatrixMode(GL_PROJECTION);
		glPushMatrix();
		glLoadIdentity();
		gluPerspective(65, aspectRatio, znear, zfar);

		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();
		glLoadIdentity();
		gluLookAt(pLight->m_pos[0], pLight->m_pos[1], pLight->m_pos[2],
			0.0, 0.0, 0.0,
			0.0, 1.0, 0.0);

		// grab modelview and transformation matrices
		glGetDoublev(GL_MODELVIEW_MATRIX, modelView);
		glGetDoublev(GL_PROJECTION_MATRIX, projection);

		glMatrixMode(GL_TEXTURE);
		glActiveTexture(GL_TEXTURE7);
		glLoadIdentity();	
		glLoadMatrixd(bias);
		
		// concatenating all matrices into one.
		glMultMatrixd (projection);
		glMultMatrixd (modelView);
		
		glActiveTexture(GL_TEXTURE0); //debug

		glMatrixMode(GL_PROJECTION);
		glPopMatrix();

		// go back to normal matrix mode
		glMatrixMode(GL_MODELVIEW);
		glPopMatrix();

		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0); // we are back in the original frame buffer
		
		glClearColor(Globals::clearColor[0], Globals::clearColor[1], Globals::clearColor[2], 1.0f);
		//glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

		// debug
		/*{
			glUseProgram(0);
			break;
		}*/
		
		// ----------------------------------------------------------------------------------------------------
		// step 3: render shadow into result
		//		
		m_shaderShadowMap->use(); // use the shadow map shader in this render pass (with m_shadowMapColor texture handle as input texture)

		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA, /*GL_DST_ALPHA*/ GL_ONE_MINUS_SRC_ALPHA); // blend results with previous results

		// tex unit 0
		glActiveTexture(GL_TEXTURE0);
		glBindTexture(GL_TEXTURE_2D, m_shadowMapColor);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_MODE, GL_NONE);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_FUNC, GL_ALWAYS);

		// render scene again, using the shadow map shader
		pScene->drawMeshes();
		
		glDisable(GL_BLEND);

		glUseProgram(0);
		//break;
	}	
}
